#define _GNU_SOURCE
#define __USE_GNU

#include <linux/module.h>
#include <linux/kernel.h>  /* Needed for KERN_ALERT */

#include <linux/sched.h> /* Need for current_thread_info() */

#include "../include/defines.h"
#include "hook.h"

//Hook the General Protection Fault handler (0x0D)
#define INTERRUPT_VECTOR_TO_HOOK 0x0D

//Opcode for RDTSC : 0F 31 => 31 OF in little endian
#define RDTSC_OPCODE 0x310F

//Size of RDTSC instruction
#define RDTSC_SIZE   2

//Count the number of hooked interrupts
volatile unsigned int nb_interrupts = 0;

//Hook this PID
volatile unsigned int pid_to_hook = 0;


//Interrupt handlers
u_int32 old_int_handler, new_int_handler2;


//If IDT_ADDRESS is given, use it, otherwise use a hardcoded value.
#ifndef IDT_ADDRESS
#  define IDT_ADDRESS 0xc06d4000
#endif



/**
 * Get a pointer to the IDT - the soft way.
 * Works perfectly in VMs, but we either have to hardcode the IDT offset,
 * or read it from userland ('grep idt_table /proc/kallsyms' or 'grep idt_table /boot/System.map').
 */
P_IDTENTRY_ST GetIDTSoft (void)
{
   P_IDTENTRY_ST pIDT = 0;

   pIDT = (P_IDTENTRY_ST) IDT_ADDRESS;

   return pIDT;
}


/**
 * Get a pointer to the IDT - the hard way
 * Doesn't work in VirtualBox because sidt is not handled correctly.
 */
P_IDTENTRY_ST GetIDTHard (void)
{
   P_IDTENTRY_ST pIDT = 0;

   asm volatile ("sidt %0\t\n" : "=m"(pIDT));

   return pIDT;
}

/**
 * Get CR4 value
 */
u_int32 GetCR4 (void)
{
   u_int32 res = 0;

   asm volatile (
         "push %%eax\t\n"
         "mov %%cr4, %%eax\t\n"
         "mov %%eax, %0\t\n"
         "pop %%eax\t\n"
         : "=m"(res));

   return res;
}

/**
 * Set CR4 value
 */
void SetCR4 (u_int32 _new_cr4)
{
   asm volatile(
         "push %%eax\t\n"
         "mov %0, %%eax\t\n"
         "mov %%eax, %%cr4\t\n"
         "pop %%eax\t\n"
   : : "m" (_new_cr4));
}

/**
 * Enable userland calls to RDTSC
 */
void EnableUserRDTSC (void)
{
  SetCR4(GetCR4() & ~FLAG_DISABLE_USER_RDTSC);
}

/**
 * Disable userland calls to RDTSC
 */
void DisableUserRDTSC (void)
{
  SetCR4(GetCR4() | FLAG_DISABLE_USER_RDTSC);
}


/**
 * Return current PID
 */
unsigned int GetCurrentPID (void)
{
   // !!! The 'current' macro doesn't work in interrupt context !
   // !!! We have to use current_thread_info()->task instead
   return current_thread_info()->task->pid;
}


/**
 * Hook
 */
void Hook ()
{
   //Get the IDT address (all CPUS use the same)
   P_IDTENTRY_ST pIDT = GetIDTSoft();
   
   printk(KERN_INFO "interrupt_handler = %08x\n", (u_int32) interrupt_handler);

   //Hook interrupt handler
   HookOneIDT(pIDT, INTERRUPT_VECTOR_TO_HOOK, &old_int_handler, (u_int32) interrupt_handler);

   //Hook RDTSC
   //DisableUserRDTSC();
   //on_each_cpu(DisableUserRDTSC, 0, 0);
}

/**
 * Unhook
 */
void UnHook ()
{
   //Unhook RDTSC
   //on_each_cpu(EnableUserRDTSC, 0, 0);
   //EnableUserRDTSC();

   //Unhook interrupt handler
   HookOneIDT(GetIDTSoft(), INTERRUPT_VECTOR_TO_HOOK, &new_int_handler2, old_int_handler);
}


/**
 * Hook an interrupt handler
 */
void HookOneIDT (P_IDTENTRY_ST _p_IDT, u_int32 _interrupt_number, u_int32* _old_address, u_int32 _new_address)
{

   printk(KERN_INFO "HookOneIDT => before hook, _old_address %x, *_old_address = %x\n", (u_int32) _old_address, *_old_address);

   asm("cli\n\t");

   *_old_address = ((_p_IDT[_interrupt_number].hi_offset << 16) | (_p_IDT[_interrupt_number].low_offset));
   _p_IDT[_interrupt_number].hi_offset = (_new_address >> 16) & 0xFFFF;
   _p_IDT[_interrupt_number].low_offset = (_new_address & 0xFFFF);

   asm("sti\n\t");

   printk(KERN_INFO "HookOneIDT => after hook, _old_address %x, *_old_address = %x\n", (u_int32) _old_address, *_old_address);
}


/**
 * Execute RDTSC instruction
 */
void RDTSC(P_RDTSC_ST pRDTSC)
{
   int eax, edx;
   asm volatile(

      "pushl %%eax\n\t"
      "pushl %%edx\n\t"

      "rdtsc\n\t"

      "mov %%eax, %0\t\n"
      "mov %%edx, %1\t\n"

      "popl %%edx\n\t"
      "popl %%eax\n\t"

      : "=m"(eax),"=m"(edx) );

   pRDTSC->eax = eax;
   pRDTSC->edx = edx;
}


/**
 * Function called by the interrupt handler.
 *  !! WARNING !! Don't call printk() inside, or the kernel will freeze !
 *
 * @param stack pointer to the stack
 * @return 0 if this is a normal #GP exception, 1 if it is due to our RDTSC hook
 */
u_int32  __attribute__((__fastcall__)) my_func_handler (P_MY_INT_STACK_ST stack)
{
   //nb_interrupts++;+
   asm volatile("lock incl nb_interrupts\n\t");
   
   //Detect if the instruction that triggered the exception is RDTSC
   if(* (u_int16*) stack->int_stack_hard_st.eip == (u_int16) RDTSC_OPCODE)
   {
      //Check who is executing RDTSC
      if(GetCurrentPID() == pid_to_hook)
      {
         //Change EAX and EDX with magic values
         stack->pusha_st.eax = 0x11223344;
         stack->pusha_st.edx = 0x55667788;
      }
      else
      {
         //Perform a normal call to RDTSC
         RDTSC_ST rdtsc;
         RDTSC(&rdtsc);

         stack->pusha_st.eax = rdtsc.eax;
         stack->pusha_st.edx = rdtsc.edx;
      }

      //Increment EIP
      stack->int_stack_hard_st.eip += RDTSC_SIZE;

      return 1;
   }
   else
   {
      return 0;
   }

}

