#!/bin/bash

# Script for loading / unloading a module, and create / delete a file to communicate to this module
#
# Syntax: load.sh -load|-unload module.ko 
#
# Notes:
# - You have to bee root in order to use this script!
# - Edit the 2 variables below to specify the name of your module and 

MODE=load

# Argument checks
usage() {
   echo "Usage: $(basename $0) [-h] [-u] [-f device_file_name [-d device_name ]] module.ko"
   echo
   echo "Load or unload module into the kernel, and optionnaly create (or delete) a device file."
   echo
   echo "Parameter:"
   echo " module.ko              Name of the module file to load"
   echo
   echo "Options:"
   echo " -u                     Unload the module instead of loading it."
   echo " -f device_file_name    Specify the name of the device to create (or delete, if used with -u), typically in /dev."
   echo " -d device_name         Use with -f. Specify the device name as it appears in /proc/devices (the one passed to alloc_chrdev_region())."
   echo "                        If not specified, use the name of the specified module file (without .ko)."
   echo " -h                     Display this help"
   echo
   exit
}


# Parse options
while getopts “hud:f:” OPTION
do
   case $OPTION in
      h)
         usage
         exit
         ;;
      u)
         MODE=unload
         ;;
      f)
         DEVICE_FILE_NAME=$OPTARG
         ;;
      d)
         DEVICE_NAME=$OPTARG
         ;;
      ?)
         usage
         exit
         ;;
   esac
done
shift `expr $OPTIND - 1`


if [ ! $# -eq 1 ]; then
   echo "Error: You must specify a module file."
   usage
   exit 1
fi

if [ ! -f "$1" ];then
   echo "Error: file '$1' not found"
   exit 1
fi

MODULE_FILE=$1


# Performing the actions
if [ $MODE = "load" ]; then

   # Load the module
   insmod $MODULE_FILE

   if [ $? -ne 0 ];then
      exit 1
   fi

   echo "Module loaded."

   # If specified, create a device file
   if [ ! -z "$DEVICE_FILE_NAME" ]; then

      # If the device name is not specified, try to guess
      if [ -z "$DEVICE_NAME" ]; then
         DEVICE_NAME=$(basename $MODULE_FILE .ko)
      fi

      # Get the major of the registred device
      MAJOR=$(grep $DEVICE_NAME /proc/devices | cut -f1 -d' ')

      if [ -z "$MAJOR" ]; then
         echo "Error: Unable to find $DEVICE_NAME in /proc/devices."
         exit 1
      fi

      # Create a file to interact with the driver
      mknod $DEVICE_FILE_NAME c $MAJOR 0 && echo "$DEVICE_FILE_NAME created ($MAJOR, 0)."
   fi

else

   # If specified, remove the device
   if [ ! -z "$DEVICE_FILE_NAME" ]; then
      rm $DEVICE_FILE_NAME && echo "$DEVICE_FILE_NAME deleted."
   fi

   if [ $? -ne 0 ];then
      exit 1
   fi

   # Remove the module
   rmmod $MODULE_FILE && echo "Module unloaded."
fi
